@@ -422,6 +422,9 @@ static int source_process_msg_cb(pa_msgobject *o, int code, void *data, int64_t
                 /* and the buffering we do on the source */
                 pa_bytes_to_usec(u->source_output_blocksize, &u->source_output->source->sample_spec);
 
+            /* Add resampler delay */
+            *((int64_t*) data) += pa_resampler_get_delay_usec(u->source_output->thread_info.resampler);
+
             return 0;
 
         case PA_SOURCE_MESSAGE_SET_VOLUME_SYNCED:
@@ -1466,6 +1469,8 @@ static bool sink_input_may_move_to_cb(pa_sink_input *i, pa_sink *dest) {
 /* Called from main context. */
 static void source_output_moving_cb(pa_source_output *o, pa_source *dest) {
     struct userdata *u;
+    uint32_t idx;
+    pa_source_output *output;
 
     pa_source_output_assert_ref(o);
     pa_assert_ctl_context();
@@ -1477,6 +1482,12 @@ static void source_output_moving_cb(pa_source_output *o, pa_source *dest) {
     } else
         pa_source_set_asyncmsgq(u->source, NULL);
 
+    /* Propagate asyncmsq change to attached virtual sources */
+    PA_IDXSET_FOREACH(output, u->source->outputs, idx) {
+        if (output->destination_source && output->moving)
+            output->moving(output, u->source);
+    }
+
     if (u->source_auto_desc && dest) {
         const char *y, *z;
         pa_proplist *pl;
