@@ -43,7 +43,13 @@
 #include <pulsecore/thread.h>
 #include <pulsecore/thread-mq.h>
 #include <pulsecore/rtpoll.h>
+
+#ifdef USE_SMOOTHER_2
+#include <pulsecore/time-smoother_2.h>
+#else
 #include <pulsecore/time-smoother.h>
+#endif
+
 #include <pulsecore/strlist.h>
 
 PA_MODULE_AUTHOR("Lennart Poettering");
@@ -59,13 +65,14 @@ PA_MODULE_USAGE(
         "format=<sample format> "
         "rate=<sample rate> "
         "channels=<number of channels> "
-        "channel_map=<channel map>");
+        "channel_map=<channel map>"
+        "remix=<boolean>");
 
 #define DEFAULT_SINK_NAME "combined"
 
 #define MEMBLOCKQ_MAXLENGTH (1024*1024*16)
 
-#define DEFAULT_ADJUST_TIME_USEC (10*PA_USEC_PER_SEC)
+#define DEFAULT_ADJUST_TIME_USEC (1*PA_USEC_PER_SEC)
 
 #define BLOCK_USEC (PA_USEC_PER_MSEC * 200)
 
@@ -79,6 +86,7 @@ static const char* const valid_modargs[] = {
     "rate",
     "channels",
     "channel_map",
+    "remix",
     NULL
 };
 
@@ -116,6 +124,14 @@ struct output {
 
     /* For communication of the stream latencies to the main thread */
     pa_usec_t total_latency;
+    struct {
+        pa_usec_t timestamp;
+        pa_usec_t sink_latency;
+        size_t output_memblockq_size;
+        uint64_t receive_counter;
+    } latency_snapshot;
+
+    uint64_t receive_counter;
 
     /* For communication of the stream parameters to the sink thread */
     pa_atomic_t max_request;
@@ -152,28 +168,46 @@ struct userdata {
 
     pa_idxset* outputs; /* managed in main context */
 
+    bool remix;
+
     struct {
         PA_LLIST_HEAD(struct output, active_outputs); /* managed in IO thread context */
         pa_atomic_t running;  /* we cache that value here, so that every thread can query it cheaply */
         pa_usec_t timestamp;
         bool in_null_mode;
-        pa_smoother *smoother;
+#ifdef USE_SMOOTHER_2
+        pa_smoother_2 *smoother;
+#else
+         pa_smoother *smoother;
+#endif
         uint64_t counter;
+
+        uint64_t snapshot_counter;
+        pa_usec_t snapshot_time;
+
+        pa_usec_t render_timestamp;
     } thread_info;
 };
 
+struct sink_snapshot {
+    pa_usec_t timestamp;
+    uint64_t send_counter;
+};
+
 enum {
     SINK_MESSAGE_ADD_OUTPUT = PA_SINK_MESSAGE_MAX,
     SINK_MESSAGE_REMOVE_OUTPUT,
     SINK_MESSAGE_NEED,
     SINK_MESSAGE_UPDATE_LATENCY,
     SINK_MESSAGE_UPDATE_MAX_REQUEST,
-    SINK_MESSAGE_UPDATE_LATENCY_RANGE
+    SINK_MESSAGE_UPDATE_LATENCY_RANGE,
+    SINK_MESSAGE_GET_SNAPSHOT
 };
 
 enum {
     SINK_INPUT_MESSAGE_POST = PA_SINK_INPUT_MESSAGE_MAX,
-    SINK_INPUT_MESSAGE_SET_REQUESTED_LATENCY
+    SINK_INPUT_MESSAGE_SET_REQUESTED_LATENCY,
+    SINK_INPUT_MESSAGE_LATENCY_SNAPSHOT
 };
 
 static void output_disable(struct output *o);
@@ -181,12 +215,48 @@ static void output_enable(struct output *o);
 static void output_free(struct output *o);
 static int output_create_sink_input(struct output *o);
 
+/* rate controller, called from main context
+ * - maximum deviation from base rate is less than 1%
+ * - controller step size is limited to 2.01‰
+ * - exhibits hunting with USB or Bluetooth devices
+ */
+static uint32_t rate_controller(
+                struct output *o,
+                uint32_t base_rate, uint32_t old_rate,
+                int32_t latency_difference_usec) {
+
+    double new_rate, new_rate_1, new_rate_2;
+    double min_cycles_1, min_cycles_2;
+
+    /* Calculate next rate that is not more than 2‰ away from the last rate */
+    min_cycles_1 = (double)abs(latency_difference_usec) / o->userdata->adjust_time / 0.002 + 1;
+    new_rate_1 = old_rate + base_rate * (double)latency_difference_usec / min_cycles_1 / o->userdata->adjust_time;
+
+    /* Calculate best rate to correct the current latency offset, limit at
+     * 1% difference from base_rate */
+    min_cycles_2 = (double)abs(latency_difference_usec) / o->userdata->adjust_time / 0.01 + 1;
+    new_rate_2 = (double)base_rate * (1.0 + (double)latency_difference_usec / min_cycles_2 / o->userdata->adjust_time);
+
+    /* Choose the rate that is nearer to base_rate */
+    new_rate = new_rate_2;
+    if (abs(new_rate_1 - base_rate) < abs(new_rate_2 - base_rate))
+        new_rate = new_rate_1;
+
+    return (uint32_t)(new_rate + 0.5);
+}
+
 static void adjust_rates(struct userdata *u) {
     struct output *o;
-    pa_usec_t max_sink_latency = 0, min_total_latency = (pa_usec_t) -1, target_latency, avg_total_latency = 0;
+    struct sink_snapshot rdata;
+    pa_usec_t avg_total_latency = 0;
+    pa_usec_t target_latency = 0;
+    pa_usec_t max_sink_latency = 0;
+    pa_usec_t min_total_latency = (pa_usec_t)-1;
     uint32_t base_rate;
     uint32_t idx;
     unsigned n = 0;
+    pa_usec_t now;
+    struct output *o_max;
 
     pa_assert(u);
     pa_sink_assert_ref(u->sink);
@@ -194,70 +264,97 @@ static void adjust_rates(struct userdata *u) {
     if (pa_idxset_size(u->outputs) <= 0)
         return;
 
-    if (!PA_SINK_IS_OPENED(u->sink->state))
+    if (u->sink->state != PA_SINK_RUNNING)
+        return;
+
+    /* Get sink snapshot */
+    pa_asyncmsgq_send(u->sink->asyncmsgq, PA_MSGOBJECT(u->sink), SINK_MESSAGE_GET_SNAPSHOT, &rdata, 0, NULL);
+
+    /* The sink snapshot time is the time when the last data was rendered.
+     * Latency is calculated for that point in time. */
+    now = rdata.timestamp;
+
+    /* Sink snapshot is not yet valid. */
+    if (!now)
         return;
 
     PA_IDXSET_FOREACH(o, u->outputs, idx) {
-        pa_usec_t sink_latency;
+        pa_usec_t snapshot_latency;
+        int64_t time_difference;
 
         if (!o->sink_input || !PA_SINK_IS_OPENED(o->sink->state))
             continue;
 
-        o->total_latency = pa_sink_input_get_latency(o->sink_input, &sink_latency);
-        o->total_latency += sink_latency;
+        /* The difference may become negative, because it is probable, that the last
+         * render time was before the sink input snapshot. In this case, the sink
+         * had some more latency at the render time, so subtracting the value still
+         * gives the right result. */
+        time_difference = (int64_t)now - (int64_t)o->latency_snapshot.timestamp;
 
-        if (sink_latency > max_sink_latency)
-            max_sink_latency = sink_latency;
+        /* Latency at sink snapshot time is sink input snapshot latency minus time
+         * passed between the two snapshots. */
+        snapshot_latency = o->latency_snapshot.sink_latency
+                           + pa_bytes_to_usec(o->latency_snapshot.output_memblockq_size, &o->sink_input->sample_spec)
+                           - time_difference;
 
-        if (min_total_latency == (pa_usec_t) -1 || o->total_latency < min_total_latency)
+        /* Add the data that was sent between taking the sink input snapshot
+         * and the sink snapshot. */
+        snapshot_latency += pa_bytes_to_usec(rdata.send_counter - o->latency_snapshot.receive_counter, &o->sink_input->sample_spec);
+
+        /* This is the current combined latency of the slave sink and the related
+         * memblockq at the time of the sink snapshot. */
+        o->total_latency = snapshot_latency;
+        avg_total_latency += snapshot_latency;
+
+        /* Get max_sink_latency and min_total_latency for target selection. */
+        if (min_total_latency == (pa_usec_t)-1 || o->total_latency < min_total_latency)
             min_total_latency = o->total_latency;
 
-        avg_total_latency += o->total_latency;
-        n++;
+        if (o->latency_snapshot.sink_latency > max_sink_latency) {
+            max_sink_latency = o->latency_snapshot.sink_latency;
+            o_max = o;
+        }
 
-        pa_log_debug("[%s] total=%0.2fms sink=%0.2fms ", o->sink->name, (double) o->total_latency / PA_USEC_PER_MSEC, (double) sink_latency / PA_USEC_PER_MSEC);
+        /* Debug output */
+        pa_log_debug("[%s] Snapshot sink latency = %0.2fms, total snapshot latency = %0.2fms", o->sink->name, (double) o->latency_snapshot.sink_latency / PA_USEC_PER_MSEC, (double) snapshot_latency / PA_USEC_PER_MSEC);
 
         if (o->total_latency > 10*PA_USEC_PER_SEC)
             pa_log_warn("[%s] Total latency of output is very high (%0.2fms), most likely the audio timing in one of your drivers is broken.", o->sink->name, (double) o->total_latency / PA_USEC_PER_MSEC);
+
+        n++;
     }
 
+    /* If there is no valid output there is nothing to do. */
     if (min_total_latency == (pa_usec_t) -1)
         return;
 
     avg_total_latency /= n;
 
-    target_latency = PA_MAX(max_sink_latency, min_total_latency);
+    /* The target selection ensures, that at least one of the
+     * sinks will use the base rate and all other sinks are set
+     * relative to it. */
+    if (max_sink_latency > min_total_latency)
+        target_latency = o_max->total_latency;
+    else
+        target_latency = min_total_latency;
 
     pa_log_info("[%s] avg total latency is %0.2f msec.", u->sink->name, (double) avg_total_latency / PA_USEC_PER_MSEC);
-    pa_log_info("[%s] target latency is %0.2f msec.", u->sink->name, (double) target_latency / PA_USEC_PER_MSEC);
+    pa_log_info("[%s] target latency for all slaves is %0.2f msec.", u->sink->name, (double) target_latency / PA_USEC_PER_MSEC);
 
     base_rate = u->sink->sample_spec.rate;
 
+    /* Calculate and set rates for the sink inputs. */
     PA_IDXSET_FOREACH(o, u->outputs, idx) {
-        uint32_t new_rate = base_rate;
-        uint32_t current_rate;
+        uint32_t new_rate;
+        int32_t latency_difference;
 
         if (!o->sink_input || !PA_SINK_IS_OPENED(o->sink->state))
             continue;
 
-        current_rate = o->sink_input->sample_spec.rate;
+        latency_difference = (int64_t)o->total_latency - (int64_t)target_latency;
+        new_rate = rate_controller(o, base_rate, o->sink_input->sample_spec.rate, latency_difference);
 
-        if (o->total_latency != target_latency)
-            new_rate += (uint32_t) (((double) o->total_latency - (double) target_latency) / (double) u->adjust_time * (double) new_rate);
-
-        if (new_rate < (uint32_t) (base_rate*0.8) || new_rate > (uint32_t) (base_rate*1.25)) {
-            pa_log_warn("[%s] sample rates too different, not adjusting (%u vs. %u).", o->sink_input->sink->name, base_rate, new_rate);
-            new_rate = base_rate;
-        } else {
-            if (base_rate < new_rate + 20 && new_rate < base_rate + 20)
-              new_rate = base_rate;
-            /* Do the adjustment in small steps; 2‰ can be considered inaudible */
-            if (new_rate < (uint32_t) (current_rate*0.998) || new_rate > (uint32_t) (current_rate*1.002)) {
-                pa_log_info("[%s] new rate of %u Hz not within 2‰ of %u Hz, forcing smaller adjustment", o->sink_input->sink->name, new_rate, current_rate);
-                new_rate = PA_CLAMP(new_rate, (uint32_t) (current_rate*0.998), (uint32_t) (current_rate*1.002));
-            }
-            pa_log_info("[%s] new rate is %u Hz; ratio is %0.3f; latency is %0.2f msec.", o->sink_input->sink->name, new_rate, (double) new_rate / base_rate, (double) o->total_latency / PA_USEC_PER_MSEC);
-        }
+        pa_log_info("[%s] new rate is %u Hz; ratio is %0.3f.", o->sink_input->sink->name, new_rate, (double) new_rate / base_rate);
         pa_sink_input_set_rate(o->sink_input, new_rate);
     }
 
@@ -271,13 +368,22 @@ static void time_callback(pa_mainloop_api *a, pa_time_event *e, const struct tim
     pa_assert(a);
     pa_assert(u->time_event == e);
 
-    adjust_rates(u);
-
     if (u->sink->state == PA_SINK_SUSPENDED) {
         u->core->mainloop->time_free(e);
         u->time_event = NULL;
-    } else
+    } else {
+        struct output *o;
+        uint32_t idx;
+
         pa_core_rttime_restart(u->core, e, pa_rtclock_now() + u->adjust_time);
+
+        /* Get latency snapshots */
+        PA_IDXSET_FOREACH(o, u->outputs, idx) {
+            pa_asyncmsgq_send(o->control_inq, PA_MSGOBJECT(o->sink_input), SINK_INPUT_MESSAGE_LATENCY_SNAPSHOT, NULL, 0, NULL);
+        }
+
+    }
+    adjust_rates(u);
 }
 
 static void process_render_null(struct userdata *u, pa_usec_t now) {
@@ -308,8 +414,13 @@ static void process_render_null(struct userdata *u, pa_usec_t now) {
 
 /*     pa_log_debug("Ate in sum %lu bytes (of %lu)", (unsigned long) ate, (unsigned long) nbytes); */
 
-    pa_smoother_put(u->thread_info.smoother, now,
-                    pa_bytes_to_usec(u->thread_info.counter, &u->sink->sample_spec) - (u->thread_info.timestamp - now));
+#ifdef USE_SMOOTHER_2
+    pa_smoother_2_put(u->thread_info.smoother, now,
+                    u->thread_info.counter - pa_usec_to_bytes(u->thread_info.timestamp - now, &u->sink->sample_spec));
+#else
+     pa_smoother_put(u->thread_info.smoother, now,
+                     pa_bytes_to_usec(u->thread_info.counter, &u->sink->sample_spec) - (u->thread_info.timestamp - now));
+#endif
 }
 
 static void thread_func(void *userdata) {
@@ -387,7 +498,10 @@ static void render_memblock(struct userdata *u, struct output *o, size_t length)
     while (pa_asyncmsgq_process_one(o->audio_inq) > 0)
         ;
 
-    /* Ok, now let's prepare some data if we really have to */
+    /* Ok, now let's prepare some data if we really have to. Save the
+     * the time for latency calculations. */
+    u->thread_info.render_timestamp = pa_rtclock_now();
+
     while (!pa_memblockq_is_readable(o->memblockq)) {
         struct output *j;
         pa_memchunk chunk;
@@ -396,6 +510,7 @@ static void render_memblock(struct userdata *u, struct output *o, size_t length)
         pa_sink_render(u->sink, length, &chunk);
 
         u->thread_info.counter += chunk.length;
+        o->receive_counter += chunk.length;
 
         /* OK, let's send this data to the other threads */
         PA_LLIST_FOREACH(j, u->thread_info.active_outputs) {
@@ -630,9 +745,10 @@ static int sink_input_process_msg(pa_msgobject *obj, int code, void *data, int64
 
         case SINK_INPUT_MESSAGE_POST:
 
-            if (PA_SINK_IS_OPENED(o->sink_input->sink->thread_info.state))
+            if (o->sink_input->sink->thread_info.state == PA_SINK_RUNNING) {
                 pa_memblockq_push_align(o->memblockq, chunk);
-            else
+                o->receive_counter += chunk->length;
+            } else
                 pa_memblockq_flush_write(o->memblockq, true);
 
             return 0;
@@ -644,6 +760,24 @@ static int sink_input_process_msg(pa_msgobject *obj, int code, void *data, int64
 
             return 0;
         }
+
+        case SINK_INPUT_MESSAGE_LATENCY_SNAPSHOT: {
+            size_t length;
+
+            length = pa_memblockq_get_length(o->sink_input->thread_info.render_memblockq);
+
+            o->latency_snapshot.output_memblockq_size = pa_memblockq_get_length(o->memblockq);
+
+            /* Add content of memblockq's to sink latency */
+            o->latency_snapshot.sink_latency = pa_sink_get_latency_within_thread(o->sink, true) +
+                                               pa_bytes_to_usec(length, &o->sink->sample_spec);
+
+            o->latency_snapshot.timestamp = pa_rtclock_now();
+
+            o->latency_snapshot.receive_counter = o->receive_counter;
+
+            return 0;
+        }
     }
 
     return pa_sink_input_process_msg(obj, code, data, offset, chunk);
@@ -674,9 +808,6 @@ static void unsuspend(struct userdata *u) {
     PA_IDXSET_FOREACH(o, u->outputs, idx)
         output_enable(o);
 
-    if (!u->time_event && u->adjust_time > 0)
-        u->time_event = pa_core_rttime_new(u->core, pa_rtclock_now() + u->adjust_time, time_callback, u);
-
     pa_log_info("Resumed successfully...");
 }
 
@@ -708,6 +839,13 @@ static int sink_set_state_in_main_thread_cb(pa_sink *sink, pa_sink_state_t state
             if (u->sink->state == PA_SINK_SUSPENDED)
                 unsuspend(u);
 
+            /* The first smoother update should be done early, otherwise the smoother will
+             * not be aware of the slave sink latencies and report far too small values.
+             * This is especially important if after an unsuspend the sink runs on a different
+             * latency than before. */
+            if (state == PA_SINK_RUNNING && !u->time_event && u->adjust_time > 0)
+                u->time_event = pa_core_rttime_new(u->core, pa_rtclock_now() + pa_sink_get_requested_latency(u->sink), time_callback, u);
+
             break;
 
         case PA_SINK_UNLINKED:
@@ -735,10 +873,17 @@ static int sink_set_state_in_io_thread_cb(pa_sink *s, pa_sink_state_t new_state,
     running = new_state == PA_SINK_RUNNING;
     pa_atomic_store(&u->thread_info.running, running);
 
-    if (running)
+    if (running) {
+        u->thread_info.render_timestamp = 0;
+#ifdef USE_SMOOTHER_2
+        pa_smoother_2_resume(u->thread_info.smoother, pa_rtclock_now());
+    } else
+        pa_smoother_2_pause(u->thread_info.smoother, pa_rtclock_now());
+#else
         pa_smoother_resume(u->thread_info.smoother, pa_rtclock_now(), true);
-    else
+    } else
         pa_smoother_pause(u->thread_info.smoother, pa_rtclock_now());
+#endif
 
     return 0;
 }
@@ -830,6 +975,7 @@ static void output_add_within_thread(struct output *o) {
             o->userdata->rtpoll,
             PA_RTPOLL_NORMAL,
             o->control_inq);
+    o->receive_counter = o->userdata->thread_info.counter;
 }
 
 /* Called from thread context of the io thread */
@@ -885,15 +1031,20 @@ static int sink_process_msg(pa_msgobject *o, int code, void *data, int64_t offse
     switch (code) {
 
         case PA_SINK_MESSAGE_GET_LATENCY: {
-            pa_usec_t x, y, c;
             int64_t *delay = data;
 
+#ifdef USE_SMOOTHER_2
+            *delay = pa_smoother_2_get_delay(u->thread_info.smoother, pa_rtclock_now(), u->thread_info.counter);
+#else
+            pa_usec_t x, y, c;
+
             x = pa_rtclock_now();
             y = pa_smoother_get(u->thread_info.smoother, x);
 
             c = pa_bytes_to_usec(u->thread_info.counter, &u->sink->sample_spec);
 
             *delay = (int64_t)c - y;
+#endif
 
             return 0;
         }
@@ -915,10 +1066,19 @@ static int sink_process_msg(pa_msgobject *o, int code, void *data, int64_t offse
             return 0;
 
         case SINK_MESSAGE_UPDATE_LATENCY: {
+#ifdef USE_SMOOTHER_2
+            size_t latency;
+
+            latency = pa_usec_to_bytes((pa_usec_t)offset,  &u->sink->sample_spec);
+            pa_smoother_2_put(u->thread_info.smoother, pa_rtclock_now(), (int64_t)u->thread_info.counter - latency);
+#else
             pa_usec_t x, y, latency = (pa_usec_t) offset;
 
-            x = pa_rtclock_now();
-            y = pa_bytes_to_usec(u->thread_info.counter, &u->sink->sample_spec);
+            /* It may be possible that thread_info.counter has been increased
+             * since we took the snapshot. Therefore we have to use the snapshot
+             * time and counter instead of the current values. */
+            x = u->thread_info.snapshot_time;
+            y = pa_bytes_to_usec(u->thread_info.snapshot_counter, &u->sink->sample_spec);
 
             if (y > latency)
                 y -= latency;
@@ -926,6 +1086,18 @@ static int sink_process_msg(pa_msgobject *o, int code, void *data, int64_t offse
                 y = 0;
 
             pa_smoother_put(u->thread_info.smoother, x, y);
+#endif
+            return 0;
+        }
+
+        case SINK_MESSAGE_GET_SNAPSHOT: {
+            struct sink_snapshot *rdata = data;
+
+            rdata->timestamp = u->thread_info.render_timestamp;
+            rdata->send_counter = u->thread_info.counter;
+            u->thread_info.snapshot_counter = u->thread_info.counter;
+            u->thread_info.snapshot_time = u->thread_info.render_timestamp;
+
             return 0;
         }
 
@@ -998,6 +1170,10 @@ static int output_create_sink_input(struct output *o) {
     data.module = u->module;
     data.resample_method = u->resample_method;
     data.flags = PA_SINK_INPUT_VARIABLE_RATE|PA_SINK_INPUT_DONT_MOVE|PA_SINK_INPUT_NO_CREATE_ON_SUSPEND;
+    data.origin_sink = u->sink;
+
+    if (!u->remix)
+        data.flags |= PA_SINK_INPUT_NO_REMIX;
 
     pa_sink_input_new(&o->sink_input, u->core, &data);
 
@@ -1287,7 +1463,7 @@ int pa__init(pa_module*m) {
     struct userdata *u;
     pa_modargs *ma = NULL;
     const char *slaves, *rm;
-    int resample_method = PA_RESAMPLER_TRIVIAL;
+    int resample_method;
     pa_sample_spec ss;
     pa_channel_map map;
     struct output *o;
@@ -1303,6 +1479,7 @@ int pa__init(pa_module*m) {
         goto fail;
     }
 
+    resample_method = m->core->resample_method;
     if ((rm = pa_modargs_get_value(ma, "resample_method", NULL))) {
         if ((resample_method = pa_parse_resample_method(rm)) < 0) {
             pa_log("invalid resample method '%s'", rm);
@@ -1320,8 +1497,15 @@ int pa__init(pa_module*m) {
         goto fail;
     }
 
+    u->remix = !m->core->disable_remixing;
+    if (pa_modargs_get_value_boolean(ma, "remix", &u->remix) < 0) {
+        pa_log("Invalid boolean remix parameter");
+        goto fail;
+    }
+
     u->resample_method = resample_method;
     u->outputs = pa_idxset_new(NULL, NULL);
+#ifndef USE_SMOOTHER_2
     u->thread_info.smoother = pa_smoother_new(
             PA_USEC_PER_SEC,
             PA_USEC_PER_SEC*2,
@@ -1330,6 +1514,7 @@ int pa__init(pa_module*m) {
             10,
             pa_rtclock_now(),
             true);
+#endif
 
     adjust_time_sec = DEFAULT_ADJUST_TIME_USEC / PA_USEC_PER_SEC;
     if (pa_modargs_get_value_u32(ma, "adjust_time", &adjust_time_sec) < 0) {
@@ -1436,6 +1621,11 @@ int pa__init(pa_module*m) {
         goto fail;
     }
 
+#ifdef USE_SMOOTHER_2
+    /* The smoother window size needs to be larger than the time between updates */
+    u->thread_info.smoother = pa_smoother_2_new(u->adjust_time + 5*PA_USEC_PER_SEC, pa_rtclock_now(), pa_frame_size(&u->sink->sample_spec), u->sink->sample_spec.rate);
+#endif
+
     u->sink->parent.process_msg = sink_process_msg;
     u->sink->set_state_in_main_thread = sink_set_state_in_main_thread_cb;
     u->sink->set_state_in_io_thread = sink_set_state_in_io_thread_cb;
@@ -1505,6 +1695,8 @@ int pa__init(pa_module*m) {
     u->sink_unlink_slot = pa_hook_connect(&m->core->hooks[PA_CORE_HOOK_SINK_UNLINK], PA_HOOK_EARLY, (pa_hook_cb_t) sink_unlink_hook_cb, u);
     u->sink_state_changed_slot = pa_hook_connect(&m->core->hooks[PA_CORE_HOOK_SINK_STATE_CHANGED], PA_HOOK_NORMAL, (pa_hook_cb_t) sink_state_changed_hook_cb, u);
 
+    u->thread_info.render_timestamp = 0;
+
     if (!(u->thread = pa_thread_new("combine", thread_func, u))) {
         pa_log("Failed to create thread.");
         goto fail;
@@ -1575,7 +1767,11 @@ void pa__done(pa_module*m) {
         u->core->mainloop->time_free(u->time_event);
 
     if (u->thread_info.smoother)
+#ifdef USE_SMOOTHER_2
+        pa_smoother_2_free(u->thread_info.smoother);
+#else
         pa_smoother_free(u->thread_info.smoother);
+#endif
 
     pa_xfree(u);
 }
